import torch
import numpy as np
from .training_utils import get_dataloader, train_model, basic_train_loop

class SimpleSklearnInterface:
    def __init__(self, network, batch_size=128, epochs=10, lr=3e-4, device="cuda:0", loss=torch.nn.CrossEntropyLoss()):
        self.network = network.to(device)
        self.optimizer = torch.optim.Adam(network.parameters(), lr=lr)
        self.criterion = loss
        self.batch_size = batch_size
        self.epochs = epochs
        self.device = device

    def fit(self, X, y):
        try:
            X = torch.from_numpy(X).to(torch.float32)
            y = torch.from_numpy(y).to(torch.long)
        except TypeError:
            pass
        train_data = get_dataloader(X, y, batch_size=self.batch_size)
        basic_train_loop(self.network, self.optimizer, self.criterion, train_data, self.epochs, self.device)

    def predict(self, X):
        res = []
        batch_size = 32
        self.network.eval()
        try:
            X = torch.from_numpy(X).to(torch.float32)
        except TypeError:
            pass

        for i in range(0, len(X), batch_size):
            predictions = (
                self.network(
                    X[i : i + batch_size].to(torch.float32).to(self.device)
                )
                .cpu()
                .detach()
                .numpy()
            )
            if predictions.shape[1] > 1:
                predictions = np.argmax(predictions, axis=1)
            else:
                predictions = np.round(predictions).astype(int)
            res.append(predictions)
        return np.concatenate(res)

    def predict_proba(self, X):
        res = []
        batch_size = 32
        self.network.eval()
        
        try:
            X = torch.from_numpy(X).to(torch.float32)
        except TypeError:
            pass

        for i in range(0, len(X), batch_size):
            predictions = (
                self.network(
                    X[i : i + batch_size].to(torch.float32).to(self.device)
                )
                .cpu()
                .detach()
                .numpy()
            )
            res.append(predictions)
        return np.concatenate(res)
    
class HypernetworkSklearnInterface:
    def __init__(self, network, batch_size=128, epochs=10, lr=3e-4, device="cuda:0", verbose=True):
        self.network = network.to(device)
        self.optimizer = torch.optim.Adam(network.parameters(), lr=lr)
        self.criterion = torch.nn.CrossEntropyLoss()
        self.batch_size = batch_size
        self.epochs = epochs
        self.verbose = verbose

        self.batch_callback = None
        self.epoch_callback = None

    def fit(self, X, y):
        try:
            X = torch.from_numpy(X).to(torch.float32)
            y = torch.from_numpy(y).to(torch.long)
        except TypeError:
            pass
        train_data = get_dataloader(X, y, batch_size=self.batch_size)
        train_model(
            self.network,
            self.optimizer,
            self.criterion,
            train_data,
            epochs=self.epochs,
            device=self.network.device,
            batch_callback=self.batch_callback,
            epoch_callback=self.epoch_callback,
            verbose=self.verbose,
        )

    def predict(self, X):
        self.network.eval()
        res = []
        batch_size = 32
        try:
            X = torch.from_numpy(X).to(torch.float32)
        except TypeError:
            pass

        for i in range(0, len(X), batch_size):
            predictions = (
                self.network(
                    X[i : i + batch_size].to(torch.float32).to(self.network.device)
                )
                .cpu()
                .detach()
                .numpy()
            )
            predictions = np.argmax(predictions, axis=1)
            res.append(predictions)
        return np.concatenate(res)

    def predict_proba(self, X):
        self.network.eval()
        
        res = []
        batch_size = 32
        try:
            X = torch.from_numpy(X).to(torch.float32).to(self.network.device)
        except TypeError:
            pass

        for i in range(0, len(X), batch_size):
            predictions = (
                self.network(
                    X[i : i + batch_size].to(torch.float32).to(self.network.device)
                )
                .cpu()
                .detach()
                .numpy()
            )
            res.append(predictions)
        return np.concatenate(res)



from . import modules
from .hypernetwork import Hypernetwork, HypernetworkPCA, HypernetworkEmbeddings, TrainingModes, HypernetworkWithFeatureSelector
from . import training_utils
from . import mask_design
from .hypertab import HyperTabClassifier


from .hypernetwork import Hypernetwork
from .interfaces import HypernetworkSklearnInterface
import numpy as np
import torch

class HyperTabClassifier:
    def __init__(
        self,
        subsample: float = 0.5,
        device: str = "cpu",
        hidden_dims: int = 10,
        test_nodes: int = 100,
        verbose: bool = False,
        epochs: int = 50,
        lr: float = 3e-4,
        batch_size: int = 128,
    ) -> None:
        """Initialize HyperTab classifier.

        Args:
            subsample (float): ratio for subsampling the columns of the input matrix
            device (str): device to use for training
            hidden_dims (int): number of neurons in the hidden layer of the target network
            test_nodes (int): number of test nodes to use. Smaller values will result in faster training,
                but less accurate results. This number should be inverse proportional
                to the number of samples in the dataset.
            verbose (bool): whether to print training progress
            epochs (int): number of epochs to train for
            lr (float): learning rate
            batch_size (int): batch size to use for training
        """

        self.subsample: float = subsample
        self.device: str = device
        self.hidden_dims: int = hidden_dims
        self.test_nodes: int = test_nodes
        self.verbose: bool = verbose
        self.epochs: int = epochs
        self.lr: float = lr
        self.batch_size: int = batch_size
        self.interface = None

    def fit(self, X, y):
        """ Train the model on the given data.

        Args:
            X (np.ndarray): input data
            y (np.ndarray): target data
        """

        try:
            X = X.numpy()
            y = y.numpy()
        except AttributeError:
            pass

        n_unique = len(np.unique(y))
        input_dims = X.shape[1]
        target_dim = int(input_dims * self.subsample)
        target_architecture = [
            (target_dim, self.hidden_dims),
            (self.hidden_dims, n_unique),
        ]

        self.hypernet = Hypernetwork(
            input_dims=input_dims,
            target_architecture=target_architecture,
            test_nodes=self.test_nodes,
            device=self.device,
        )
        self.hypernet.to(self.device)
        interface = HypernetworkSklearnInterface(
            self.hypernet,
            batch_size=self.batch_size,
            epochs=self.epochs,
            lr=self.lr,
            device=self.device,
            verbose=self.verbose,
        )

        # cast X, y to torch
        X = torch.from_numpy(X).to(torch.float32)
        y = torch.from_numpy(y).to(torch.long)

        interface.fit(X, y)
        self.interface = interface
    
    def predict(self, X):
        """ Return the predicted value of each sample in X """
        try:
            X = X.numpy()
        except AttributeError:
            pass

        if self.interface is None:
            raise ValueError("Model not trained yet")
        
        X = torch.from_numpy(X).to(torch.float32)
        return self.interface.predict(X)

    def predict_proba(self, X):
        """ Return the predicted class probability of each sample in X """
        try:
            X = X.numpy()
        except AttributeError:
            pass

        if self.interface is None:
            raise ValueError("Model not trained yet")

        X = torch.from_numpy(X).to(torch.float32)
        return self.interface.predict_proba(X)



from dotenv import load_dotenv

import os

import torch
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import numpy as np
import torch.utils.data as data_utils
from sklearn.metrics import roc_auc_score, f1_score, precision_score, recall_score, balanced_accuracy_score
from .hypernetwork import TrainingModes

torch.set_default_dtype(torch.float32)

from tqdm import trange


def get_dataset(size=60000, masked=False, mask_no=200, mask_size=700, shared_mask=False, batch_size=32, test_batch_size=32):
    mods = [transforms.ToTensor(), 
        transforms.Normalize((0.1307,), (0.3081,)),    #mean and std of MNIST
        transforms.Lambda(lambda x: torch.flatten(x))]
    mods = transforms.Compose(mods)
    
    trainset = datasets.MNIST(root='./data/train', train=True, download=True, transform=mods)
    testset = datasets.MNIST(root='./data/test', train=False, download=True, transform=mods)
    if masked:
        trainset = MaskedDataset(trainset, mask_no, mask_size)
        testset = MaskedDataset(testset, mask_no, mask_size)
        if shared_mask:
            testset.masks = trainset.masks
        
    indices = torch.arange(size)
    trainset = data_utils.Subset(trainset, indices)
    
    trainloader = torch.utils.data.DataLoader(trainset, batch_size=batch_size,
                                          shuffle=True, num_workers=1)
    testloader = torch.utils.data.DataLoader(testset, batch_size=test_batch_size,
                                         shuffle=False, num_workers=1)
    return trainloader, testloader



class MaskedDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, masks, mask_size):
        inputs = dataset[0][0].shape[0]
        self.mask_size = mask_size
        self.dataset = dataset
        self.template = np.zeros(inputs)
        self.masks = self._create_mask(masks)
        self.masks_indices = np.random.choice(np.arange(masks), len(dataset))
        
    def _create_mask(self, count):
        masks = np.array([np.random.choice((len(self.template)), self.mask_size, False) for _ in range(count)])
        tmp = np.array([self.template.copy() for _ in range(count)])
        for i, mask in enumerate(masks):
            tmp[i, mask] = 1
        mask = torch.from_numpy(tmp).to(torch.float32)
        return mask
    
    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        image, label = self.dataset[idx]
        mask = self.masks[self.masks_indices[idx]]
        return image, label, mask

def test_model(hypernet, testloader, device='cpu', verbose=False):
    criterion = torch.nn.CrossEntropyLoss()
    correct = 0
    loss = 0
    hypernet.eval()
    for i, data in enumerate(testloader):
        try:
            images, labels, _ = data
        except ValueError:
            images, labels = data
            
        images = images.to(device)
        labels = labels.to(device)
        outputs = hypernet(images)
        _, predicted = torch.max(outputs.data, 1)
        correct += (predicted == labels).sum().item()
        loss += criterion(outputs, labels).item()

    if verbose:
        print(f"Test acc: {correct/len(testloader.dataset)*100:.2f}")
        print(f"Test loss: {loss/(i+1):.2g}")
    return correct/len(testloader.dataset)*100


def batch_predict(network, loader, device='cpu'):
    res = []
    for i, data in enumerate(loader):
        try:
            images, labels, _ = data
        except ValueError:
            images, labels = data
            
        images = images.to(device)
        labels = labels.to(device)
        outputs = network(images)
        res.append(outputs.cpu().detach().numpy())
    return np.concatenate(res)

def train_model(hypernet, 
                optimizer, 
                criterion, 
                trainloader, 
                epochs,
                batch_callback=None,
                epoch_callback=None,
                device='cpu',
                verbose=True):
    mask_idx = 0
    masks = None
    with trange(epochs, disable=not verbose) as t:
        for _ in t:
            total_loss = 0
            hypernet.train()
            y_pred = []
            y_true = []
            for num, data in enumerate(trainloader):
                try:
                    inputs, labels, _ = data
                except ValueError:
                    inputs, labels = data
                y_true.extend(labels.tolist())
                inputs = inputs.to(device)
                labels = labels.to(device)

                if hypernet.mode == TrainingModes.SLOW_STEP:
                    masks = hypernet.test_mask[mask_idx].repeat(len(inputs), 1)
                    mask_idx = (mask_idx+1) % len(hypernet.test_mask)

                    optimizer.zero_grad()

                    outputs = hypernet(inputs, masks)
                    loss = criterion(outputs, labels)
                    loss.backward()
                    total_loss += loss.item()
                    optimizer.step()
                    y_pred.extend(outputs.tolist())

                elif hypernet.mode == TrainingModes.CARTHESIAN:
                    mask_order = np.arange(len(hypernet.test_mask))
                    np.random.shuffle(mask_order)
                    preds = []
                    for mask_idx in mask_order:
                        masks = hypernet.test_mask[mask_idx].repeat(len(inputs), 1)

                        optimizer.zero_grad()

                        outputs = hypernet(inputs, masks)
                        loss = criterion(outputs, labels)
                        loss.backward()
                        total_loss += loss.item()
                        
                        optimizer.step()
                        preds.append(outputs.tolist())
                    preds = np.mean(preds, axis=0).tolist()
                    y_pred.extend(preds)

                if batch_callback is not None:
                    batch_callback({
                        "batch_loss": loss.item()
                    })
            total_loss /= (num+1)
            if epoch_callback is not None:
                epoch_callback({
                    "total_loss": total_loss,
                    "bacc": balanced_accuracy_score(y_true, np.argmax(y_pred, axis=-1))
                })
    return total_loss

def basic_train_loop(network, optimizer, criterion, trainloader, epochs, device="cpu"):
    for _ in range(epochs):
        network.train()
        total_loss = 0
        for i, data in enumerate(trainloader):
            inputs, labels = data
            inputs = inputs.to(device)
            labels = labels.to(device)

            optimizer.zero_grad()

            outputs = network(inputs)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()
            total_loss += loss.item()
    return total_loss/(i+1)

class GenericDataset(torch.utils.data.Dataset):
    def __init__(self, data, samples_no: int=None):
        samples = samples_no or len(data[0])
        self.indices = np.arange(samples)
        self.index = 0
        self.max_samples = samples
        self.data_x = data[0].to(torch.float32)
        self.data_y = data[1]

    def __getitem__(self, idx):
        return self.data_x[idx], self.data_y[idx]
    
    def __len__(self):
        return self.max_samples
    
def get_dataloader(X, y, size=None, batch_size=32, shuffle=True):
    train_dataset = GenericDataset((X, y))
    trainloader = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, num_workers=1, shuffle=shuffle)
    
    return trainloader


import torch
import numpy as np
from os.path import exists


def feature_variances(data):
    trainset, _ = data
    stacked_set = torch.stack([x for x, _ in trainset])

    feature_variances = torch.var(stacked_set, dim=0)

    return feature_variances


class VarianceWithThresholdMasksSelector():
    def __init__(self, data, mask_size, threshold=0.2):

        self.mask_size = mask_size

        trainset, _ = data
        stacked_set = torch.stack([x for x, _ in trainset])

        feature_variances = torch.var(stacked_set, dim=0)
        self.feat_above_thresh = (feature_variances > threshold).nonzero()[:, 0]


    def _random_choice_mask(self, mask_size):
        tmp = np.random.choice(self.feat_above_thresh, mask_size, replace=False)
        tmp.sort()

        return tmp

    def __call__(self, model, count):
        print(model)
        print(self)

        tmp = np.array([model.template.copy() for _ in range(count)])

        for i in range(count):
            mask = self._random_choice_mask(model.mask_size)
            tmp[i, mask] = 1

        masks = torch.from_numpy(tmp).to(torch.float32).to(model.device)
        print('VarianceWithThresholdMasksSelector masks shape', masks.shape)
        return masks



class VarianceWithSoftmaxMasksSelector():
    def __init__(self, feature_variances, mask_size, temp_scheduler):

        self.mask_size = mask_size

        self.feature_variances = feature_variances

        self.temp_scheduler = temp_scheduler


    def _random_choice_mask(self, mask_size):
        mask = []
        unused = np.full(self.feature_variances.shape, True)

        for i in range(mask_size):

            current_temp = self.temp_scheduler.update_temp(i)

            probs = torch.nn.functional.softmax(self.feature_variances[unused] / current_temp, dtype=torch.float32, dim=0).numpy()

            idx = np.random.choice(unused.nonzero()[0], 1, replace=False, p=probs)
            mask.append(idx[0])
            unused[idx[0]] = False

        mask = np.array(mask)
        mask.sort()

        return mask


    def __call__(self, model, count):
        tmp = np.array([model.template.copy() for _ in range(count)])

        for i in range(count):
            self.temp_scheduler.reset_to_initial()

            mask = self._random_choice_mask(model.mask_size)
            tmp[i, mask] = 1

        masks = torch.from_numpy(tmp).to(torch.float32).to(model.device)
        print('VarianceWithSoftmaxMasksSelector masks shape', masks.shape)
        return masks



class IncByOneTempScheduler():
    def __init__(self, initial_temp, max_temp=20, n=20):

        self.initial_temp = initial_temp
        self.available_temps = np.linspace(initial_temp, max_temp, n)

    def reset_to_initial(self, initial=None):
        if not initial:
            initial = self.initial_temp

        self.temp = initial

    def update_temp(self, n):
        self.temp = self.available_temps[n]

        return self.temp


class HalfHalfTempScheduler():
    def __init__(self, interval1=(-11, -1), interval2=(1, 11), n=20):
        half = n//2
        self.available_temps = np.concatenate((np.linspace(*interval1, half), np.linspace(*interval2, n-half)))

    def reset_to_initial(self, initial=None):
        pass

    def update_temp(self, n):
        self.temp = self.available_temps[n]

        return self.temp

def get_feature_accuracies(dataset, seed=None, file=None):

    if file:
        if exists(file):
            return torch.load(file)
        else:
            print("Invalid path")

    trainset, _ = dataset
    stacked_train_set = torch.stack([x for x, _ in trainset])
    stacked_test_set = torch.Tensor([y for _, y in trainset])

    print('dataset shape', stacked_train_set.shape)

    res = []
    for i in tqdm(range(stacked_train_set.shape[1])):

        X = stacked_train_set[:, i].reshape(-1,1)
        y = stacked_test_set

        model = LogisticRegression(random_state=seed).fit(X, y)
        res.append(model.score(X, y))

    res_tensor = torch.tensor(res)

    if file:
        torch.save(res_tensor, file)

    return res_tensor


class AccMasksSelector():
    def __init__(self, feature_acc, mask_size, temp_scheduler):

        self.mask_size = mask_size

        self.feature_acc = feature_acc

        self.temp_scheduler = temp_scheduler


    def _random_choice_mask(self, mask_size):
        mask = []
        unused = np.full(self.feature_acc.shape, True)

        for i in range(mask_size):

            current_temp = self.temp_scheduler.update_temp(i)

            probs = torch.nn.functional.softmax(self.feature_acc[unused] / current_temp, dtype=torch.float32, dim=0).numpy()

            idx = np.random.choice(unused.nonzero()[0], 1, replace=False, p=probs)
            mask.append(idx[0])
            unused[idx[0]] = False

        mask = np.array(mask)
        mask.sort()

        return mask


    def __call__(self, model, count):
        tmp = np.array([model.template.copy() for _ in range(count)])

        for i in range(count):
            self.temp_scheduler.reset_to_initial()

            mask = self._random_choice_mask(model.mask_size)
            tmp[i, mask] = 1

        masks = torch.from_numpy(tmp).to(torch.float32).to(model.device)
        print('AccMasksSelector masks shape', masks.shape)
        return masks






from setuptools import setup, find_packages


with open('requirements.txt') as f:
    required = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="hypertab",
    author="Witold Wydmański",
    author_email="wwydmanski@gmail.com",
    version="0.1.7",
    description="HyperTab: hypernetwork for small tabular datasets",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="MIT",
    packages=find_packages(include=['hypertab', 'hypertab.*']),
    install_requires=required,
    url="https://github.com/wwydmanski/hypertab"
)



from dotenv import load_dotenv
load_dotenv()

import os

import torch
import numpy as np
import torch.nn.functional as F

torch.set_default_dtype(torch.float32)


class SimpleNetwork(torch.nn.Module):
    def __init__(self, inp_size, layers=[100], outputs=10):
        super().__init__()
        self.layers = []
        
        self.inp = torch.nn.Linear(inp_size, layers[0])
        self.output = torch.nn.Linear(layers[0], outputs)
        self.relu = torch.nn.ReLU()
        self.softmax = torch.nn.Softmax()
        self.sigmoid = torch.nn.Sigmoid()
    
    def forward(self, x):
        out = x
        out = self.inp(out)
        out = self.relu(out)
        out = self.output(out)
        return out
    
class DropoutNetwork(torch.nn.Module):
    def __init__(self, inp_size, layers=[100]):
        super().__init__()
        self.layers = []
        
        self.inp = torch.nn.Linear(inp_size, layers[0])
        self.output = torch.nn.Linear(layers[0], 10)
        self.relu = torch.nn.ReLU()
        self.softmax = torch.nn.Softmax()
        self.sigmoid = torch.nn.Sigmoid()
        self.dropout = torch.nn.Dropout(1-700/784)
    
    def forward(self, x):
        out = x
        out = self.dropout(out)
        out = self.inp(out)
        out = self.relu(out)
        out = self.output(out)
        return out

    
class InsertableNet(torch.nn.Module):
    def __init__(self, weights, shape=[(784, 10), (10, 10)]):
        super().__init__()
        self.layers = []
        self._offset = 0
        
        for layer in shape:
            _w_size = layer[0]*layer[1]
            _b_size = layer[1]
            
            _l = (weights[self._offset:self._offset+_w_size].reshape((layer[1], layer[0])),
                  weights[self._offset+_w_size:self._offset+_w_size+_b_size])
            self._offset += _w_size+_b_size

            self.layers.append(_l)
    
    def forward(self, data):
        out = data
        for layer in self.layers[:-1]:
            out = F.linear(out, layer[0], layer[1])
            out = F.relu(out)

        return F.linear(out, self.layers[-1][0], self.layers[-1][1])
    
    
class MaskedNetwork(SimpleNetwork):
    def __init__(self, input_size, mask_size, layers=[10]):
        super().__init__(mask_size, layers=layers)
        template = np.zeros(input_size)
        mask = np.random.choice(len(template), mask_size, False)
        template[mask] = 1
        self.mask = torch.from_numpy(template).to(torch.bool)
        
    def forward(self, x):
        data = x[:, self.mask]
        return super().forward(data)




import torch
import numpy as np
from .modules import InsertableNet
import enum
import torch.nn.functional as F
from sklearn.decomposition import PCA

torch.set_default_dtype(torch.float32)

class TrainingModes(enum.Enum):
    SLOW_STEP = "slow-step"
    CARTHESIAN = "carth"

class Hypernetwork(torch.nn.Module):
    def __init__(
        self,
        input_dims=None,
        architecture=None,
        target_architecture=[(20, 10), (10, 10)],
        test_nodes=100,
        mode=TrainingModes.CARTHESIAN,
        device="cuda:0",
    ):
        """ Initialize a hypernetwork.
        Args:
            target_inp_size - size of input
            out_size - size of output
            layers - list of hidden layer sizes
            test_nodes - number of test nodes
            device - device to use
        """
        super().__init__()
        if architecture is None and input_dims is not None:
            architecture = torch.nn.Sequential(torch.nn.Linear(input_dims, 128), 
                        torch.nn.ReLU(),
                        torch.nn.Linear(128, 64),
                        torch.nn.ReLU(),
                        torch.nn.Linear(64, 64)
                       )
        elif architecture is None and input_dims is None:
            raise ValueError("Either `input_dims` or `architecture` is required.")
        self.target_outsize = target_architecture[-1][-1]
        self.mask_size = target_architecture[0][0]
        self.target_architecture = target_architecture
        self.device = device
        self.mode = mode

        self.out_size = self.calculate_outdim(target_architecture)

        self.model = architecture.to('cpu')
        gen = self.model.parameters()
        self.input_size = next(gen).size()[1]
        
        out_dim = self.model(torch.rand(1, self.input_size)).shape
        
        output_layer = torch.nn.Linear(out_dim[1], self.out_size)
        self.model.add_module("output_layer", output_layer)
        self.model = self.model.to(device)
        
        self.dropout = torch.nn.Dropout()

        self.relu = torch.relu
        self.template = np.zeros(self.input_size)
        self.test_nodes = test_nodes
        self.test_mask = self._create_mask(test_nodes)

        self._retrained = True
        self._test_nets = None
        
    def forward(self, data, mask=None):
        """Get a hypernet prediction.
        During training we use a single target network per sample.
        During eval, we create a network for each test mask and average their results

        Args:
            data - prediction input
            mask - either None or a torch.tensor((data.shape[0], data.shape[1])).
        """
        if self.training:
            self._retrained = True
            if self.mode == TrainingModes.SLOW_STEP or self.mode == TrainingModes.CARTHESIAN:
                return self._slow_step_training(data, mask)

            if mask is None:
                mask = self._create_mask(len(data))

            return self._external_mask_training(data, mask)
        else:
            return self._ensemble_inference(data, mask)
            
    def calculate_outdim(self, architecture):
        weights = 0
        for layer in architecture:
            weights += layer[0]*layer[1]+layer[1]
        return weights

    def to(self, device):
        super().to(device)
        self.device = device
        self.test_mask = self._create_mask(self.test_nodes)
        self.model = self.model.to(device)
        return self

    def _slow_step_training(self, data, mask):
        weights = self.craft_network(mask[:1])
        mask = mask[0].to(torch.bool)
        nn = InsertableNet(
            weights[0],
            self.target_architecture,
        )

        masked_data = data[:, mask]
        res = nn(masked_data)
        return res

    def _external_mask_training(self, data, mask):
        recalculate = [True] * len(mask)
        for i in range(1, len(mask)):
            if torch.equal(mask[i - 1], mask[i]):
                recalculate[i] = False

        weights = self.craft_network(mask)
        mask = mask.to(torch.bool)

        res = torch.zeros((len(data), self.target_outsize)).to(self.device)
        for i in range(len(data)):
            if recalculate[i]:
                nn = InsertableNet(
                    weights[i],
                    self.target_architecture,
                )
            masked_data = data[i, mask[i]]
            res[i] = nn(masked_data)
        return res


    def _ensemble_inference(self, data, mask):
        if mask is None:
            mask = self.test_mask
            nets = self._get_test_nets()
        else:
            nets = self.__craft_nets(mask)
        mask = mask.to(torch.bool)

        res = torch.zeros((len(data), self.target_outsize)).to(self.device)
        for i in range(len(mask)):
            nn = nets[i]
            masked_data = data[:, mask[i]]
            res += nn(masked_data)
        res /= len(mask)
        return res

    def _get_test_nets(self):
        if self._retrained:
            nets = self.__craft_nets(self.test_mask)
            self._test_nets = nets
            self._retrained = False
        return self._test_nets

    def __craft_nets(self, mask):
        nets = []
        weights = self.craft_network(mask.to(torch.float32))
        for i in range(len(mask)):
            nn = InsertableNet(
                weights[i],
                self.target_architecture,
            )
            nets.append(nn)
        return nets

    @staticmethod
    def random_choice_noreplace2(l, n_sample, num_draw):
        '''
        l: 1-D array or list
        n_sample: sample size for each draw
        num_draw: number of draws

        Intuition: Randomly generate numbers, get the index of the smallest n_sample number for each row.
        '''
        l = np.array(l)
        return l[np.argpartition(np.random.rand(num_draw,len(l)), n_sample-1,axis=-1)[:,:n_sample]]
    
    def _create_mask(self, count):
        # masks = np.random.choice((len(self.template)), (count, self.mask_size), False)
        masks = Hypernetwork.random_choice_noreplace2(np.arange(len(self.template)), self.mask_size, count)
        tmp = np.array([self.template.copy() for _ in range(count)])
        for i, mask in enumerate(masks):
            tmp[i, mask] = 1
        mask = torch.from_numpy(tmp).to(torch.float32).to(self.device)
        return mask

    def craft_network(self, mask):
        # embs = self.embeddings(mask.nonzero().to(torch.int))[:, 1]
        # embs = embs.reshape(len(mask), -1, embs.shape[-1])
        # embs = embs.sum(1)
        
        out = self.model(mask)
        return out


class HypernetworkEmbeddings(Hypernetwork):
    def __init__(self, input_size=784, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = self.model[:-1].to("cpu")
        self.input_size = input_size
        
        gen = self.model.parameters()
        embedding_size = next(gen).size()[1]
        
        out_dim = self.model(torch.rand(1, embedding_size)).shape
        output_layer = torch.nn.Linear(out_dim[1], self.out_size)
        
        self.test_mask = self._create_mask(self.test_nodes)
        
        self.model.add_module("output_layer", output_layer)
        self.embeddings = torch.nn.Embedding(input_size, embedding_size, )
        self.template = np.zeros(self.input_size)

    def craft_network(self, mask):
        embs = self.embeddings(mask.nonzero().to(torch.int))[:, 1]
        embs = embs.reshape(len(mask), -1, embs.shape[-1])
        embs = embs.sum(1)
        
        out = self.model(embs)
        return out


class HypernetworkPCA(Hypernetwork):
    def __init__(self, input_size=784, *args, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except ValueError:
            # It happens during `self._create_mask(test_nodes)`
            # due to wrong mask size
            pass
        
        self.input_size = input_size
        self.template = np.zeros(self.input_size)
        self.model = self.model[:-1].to("cpu")
        
        gen = self.model.parameters()
        embedding_size = next(gen).size()[1]
        self.embedding_size = embedding_size
        
        out_dim = self.model(torch.rand(1, embedding_size)).shape
        output_layer = torch.nn.Linear(out_dim[1], self.out_size)
        
        self.model.add_module("output_layer", output_layer)
        self.template = np.zeros(self.input_size)
        self.test_mask = self._create_mask(self.test_nodes)
            
        self.pca = self._get_pca(self.test_mask.cpu().detach())

    def craft_network(self, mask):
        device = mask.device
        embs = self.pca.transform(mask.to("cpu"))
        embs = torch.tensor(embs).to(device).to(torch.float32)
        
        out = self.model(embs)
        return out
    
    def _get_pca(self, masks):
        pca = PCA(self.embedding_size)
        pca.fit(masks)
        return pca
    

class HypernetworkWithFeatureSelector(Hypernetwork):
    def __init__(self, input_size=784, feature_selector=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.test_mask = feature_selector(self, self.test_nodes)
        self._create_mask = lambda x: feature_selector(self, x)    
